/**
 * @author Syed Husain
 */
var signalR = require('signalr-client');
var EventHandler = require('../../util/eventhandler.js');
var RequestUtil = require('../../util/requestUtil');
var xml2js = require('xml2js');
var PMObject = require('./pmobject').PriorityMessagingObject;

function CommunicationService(config) {
	var appspaceURL = config.appspaceURL;
	var playerID = config.playerID;

	this.onReceiveUpdates = null;
	this.onReceiveError = null;

	// polling
	var pollingDuration = 20;
	var onExecutePollingEventHandler = null;
	var pollingHandler = -1;
	var pollingURL = '';
	var pollingOnSuccessCall = null;
	var pollingOnErrorCall = null;
	var previousPollingData = null;
	var pollingOnXMLParsed = null;

	// signalr
	var signalrClient = null;
	var hubName = 'priorityMessagesHub';
	var onSignalrErrorEventHandler = null;
	var onReceiveSignalrMessageEventHandler = null;
	var onSignalrConnectedEventHandler = null;
	var onSignalrTimeoutEventHandler = null;
	var checkSignalrStat = -1;
	var signalrConnected = false;

	this.start = function() {
		// console.log(appspaceURL + ' ' + playerID);
		// this.startManualPolling();
		this.initSignalrClient();
		this.startSignalRClient();
	};

	this.initSignalrClient = function() {
		onSignalrErrorEventHandler = EventHandler.CreateEventHandler(this, this.onSignalrErrorHandler);
		onSignalrConnectedEventHandler = EventHandler.CreateEventHandler(this, this.onSignalrConnectedHandler);
		onReceiveSignalrMessageEventHandler = EventHandler.CreateEventHandler(this, this.onSignalrMessageHandler);
		onSignalrTimeoutEventHandler = EventHandler.CreateEventHandler(this, this.onSignalrTimeoutHandler);
	};

	this.startSignalRClient = function() {
		var clientURL = appspaceURL + "/priority/signalr";// /start?clientguid="
		// + playerID;
		hubName = 'priorityMessagesHub';
		console.log('connecting to ' + clientURL + ' (hub:' + hubName + ')');
		if (signalrClient !== null) {
			signalrClient.end();
			signalrClient = null;
		}

		signalrClient = new signalR.client(clientURL,// signalR service URL
		[ hubName ],// hubs
		5,// reconnection timeout (in seconds)
		true// disable auto start?
		);

		signalrClient.queryString.clientguid = playerID;
		
		signalrClient.serviceHandlers.bindingError = onSignalrErrorEventHandler;
		signalrClient.serviceHandlers.onerror = onSignalrErrorEventHandler;
		signalrClient.serviceHandlers.connectFailed = onSignalrErrorEventHandler;
		signalrClient.serviceHandlers.onUnauthorized = onSignalrErrorEventHandler;
		signalrClient.serviceHandlers.connected = onSignalrConnectedEventHandler;
		signalrClient.handlers.prioritymessageshub = {
			onprioritymessage : onReceiveSignalrMessageEventHandler
		};
		checkSignalrStat = setTimeout(onSignalrTimeoutEventHandler, 20000);
		signalrClient.start();
	};

	this.onSignalrErrorHandler = function(calle, error) {
		calle.onSignalrError(error);
	};

	this.onSignalrError = function(error) {

		clearTimeout(checkSignalrStat);
		checkSignalrStat = -1;

		console.log('receive error, reverting to manual polling');
		console.log(error);

		if (signalrClient !== null) {
			signalrClient.end();
			signalrClient = null;
		}

		this.startManualPolling();
	};

	this.onSignalrConnectedHandler = function(calle, connection) {
		calle.onSignalrConnected(connection);
	};

	this.onSignalrConnected = function(connection) {
		console.log('signalR connected');
		clearTimeout(checkSignalrStat);
		checkSignalrStat = -1;
		signalrConnected = true;
		

	};

	this.onSignalrMessageHandler = function(calle, message) {
		calle.onSignalrMessage(message);
	};

	this.onSignalrMessage = function(message) {
		console.log('signalR receive message');
		console.log(message);
		var obj = JSON.parse(message);
		var pmObject = new PMObject();
		pmObject.copy(obj);
		//console.log(pmObject);
		
		if (signalrClient !== null) {
			console.log('sending acknowledgement:' + pmObject.StateGuid);
			var resp = signalrClient.hub(hubName).invoke('acknowledge',pmObject.StateGuid);
			//console.log(resp);
		}
		
		if (this.onReceiveUpdates !== null) {
			this.onReceiveUpdates(pmObject);
		}
	};

	this.onSignalrTimeoutHandler = function(calle) {
		calle.onSignalrTimeout();
	};

	this.onSignalrTimeout = function() {
		this.onSignalrError('timeout - doesn\'t receive signal');
	};

	this.startManualPolling = function() {
		onExecutePollingEventHandler = EventHandler.CreateEventHandler(this, this.executePollingHandler);
		pollingURL = appspaceURL + "/priority/RSSFeed/CampaignFeeds/" + playerID;
		pollingOnSuccessCall = EventHandler.CreateEventHandler(this, this.onPollingReceiveSuccessHandler);
		pollingOnErrorCall = EventHandler.CreateEventHandler(this, this.onPollingReceiveErrorHandler);
		pollingOnXMLParsed = EventHandler.CreateEventHandler(this, this.onPollingXMLParsedHandler);
		this.startPollingTimer();
	};

	this.startPollingTimer = function() {
		clearTimeout(pollingHandler);
		pollingHandler = setTimeout(onExecutePollingEventHandler, 1000);
	};

	this.executePollingHandler = function(calle) {
		calle.executePolling();
	};

	this.executePolling = function() {
		var d = new Date();
		if (d.getSeconds() % pollingDuration === 0) {
			this.callAppspace();
			return;
		}

		this.startPollingTimer();
	};

	this.callAppspace = function() {
		console.log('Retrieving feed from Appspace');
		var options = RequestUtil.createOptionsObject();
		options.url = pollingURL;
		options.failCallback = pollingOnErrorCall;
		options.successCallback = pollingOnSuccessCall;
		RequestUtil.makeRequest(options);
	};

	this.onPollingReceiveSuccessHandler = function(calle, res, body) {
		calle.onPollingReceiveSuccess(res, body);
	};

	this.onPollingReceiveSuccess = function(res, body) {
		if (body !== null && body !== undefined) {
			var parser = new xml2js.Parser();
			parser.parseString(body, pollingOnXMLParsed);
		} else {
			this.sendPollingError('body is null/undefined');
			this.startPollingTimer();
		}
	};

	this.onPollingXMLParsedHandler = function(calle, err, result) {
		calle.onPollingXMLParsed(err, result);
	};

	this.onPollingXMLParsed = function(err, result) {
		if (err !== null && err !== undefined) {
			console.log('error during xml parsing');
			console.log(err);
			this.sendPollingError('error during xml parsing');
			this.startPollingTimer();
			return;
		}

		if (!result || !result.rss || !result.rss.channel) {
			this.sendPollingError('result or result.rss or result.rss.channel is null/undefined');
			this.startPollingTimer();
			return;
		}

		var media = null;
		var startdate = null;
		var enddate = null;
		var mediatype = null;
		var item = result.rss.channel[0].item[0];
		var title = item.title.toString();
		var link = item.link.toString();
		var desc = item.description.toString();

		if (desc === null || desc === '' || desc === undefined) {
			this.sendPollingError('description is empty');
			this.startPollingTimer();
			return;
		}

		if (previousPollingData === desc) {
			console.log('PM already been processed');
			this.startPollingTimer();
			return;
		}

		previousPollingData = desc;
		var pmItems = desc.toString().split(';');
		var l = pmItems.length;
		for (var i = 0; i < l; i++) {
			var kvp = pmItems[i].toString().split('=');
			if (kvp === null || kvp === undefined || kvp.length < 2) {
				continue;
			}
			switch (kvp[0].toLowerCase()) {
			case 'media':
				media = kvp[1];
				break;
			case 'startdate':
				startdate = kvp[1].toString();
				break;
			case 'enddate':
				enddate = kvp[1].toString();
				break;
			case 'mediatype':
			case 'contenttype':
				mediatype = kvp[1].toString().toLowerCase().split('_').join(' ');
				break;
			}
		}
		/*
		 * console.log('received:'); console.log(startdate);
		 * console.log(enddate); console.log(media); console.log(mediatype);
		 */
		console.log('Received PM');
		var pmObject = new PMObject();
		pmObject.copy({
			StartedAt : startdate,
			EndedAt : enddate,
			Media : media,
			ContentType : mediatype
		});
		console.log(pmObject);
		if (this.onReceiveUpdates !== null) {
			this.onReceiveUpdates(pmObject);
		}
		this.startPollingTimer();
	};

	this.onPollingReceiveErrorHandler = function(calle, err) {
		calle.onPollingReceiveError(err);
	};

	this.onPollingReceiveError = function(err) {
		console.log('polling receive error');
		console.log(err.message);
		this.sendPollingError('polling request receive error');
		this.startPollingTimer();
	};

	this.sendPollingError = function(message) {
		//console.log(message);
		if (this.onReceiveError !== null) {
			this.onReceiveError(message);
		}
	};

}

exports.CommunicationService = CommunicationService;